/*
 * Copyright (C) 1997 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file client_common.c Common functions of para_client and para_audiod. */

#include <netinet/in.h>
#include <sys/socket.h>
#include <regex.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <netdb.h>

#include "para.h"
#include "error.h"
#include "list.h"
#include "sched.h"
#include "crypt.h"
#include "net.h"
#include "fd.h"
#include "sideband.h"
#include "string.h"
#include "client.cmdline.h"
#include "client.h"
#include "buffer_tree.h"
#include "version.h"
#include "ggo.h"

/** The size of the receiving buffer. */
#define CLIENT_BUFSIZE 4000

/**
 * Close the connection to para_server and free all resources.
 *
 * \param ct Pointer to the client data.
 *
 * \sa \ref client_open().
 */
void client_close(struct client_task *ct)
{
	if (!ct)
		return;
	free(ct->user);
	free(ct->config_file);
	free(ct->key_file);
	client_cmdline_parser_free(&ct->conf);
	free(ct->challenge_hash);
	sb_free(ct->sbc[0]);
	sb_free(ct->sbc[1]);
	free(ct);
}

/*
 * The preselect hook for server commands.
 *
 * The task pointer must contain a pointer to the initialized client data
 * structure as it is returned by client_open().
 *
 * This function checks the state of the connection and adds the file descriptor
 * of the connection to the read or write fd set of s accordingly.
 */
static void client_pre_select(struct sched *s, void *context)
{
	int ret;
	struct client_task *ct = context;

	if (ct->scc.fd < 0)
		return;
	switch (ct->status) {
	case CL_CONNECTED:
	case CL_SENT_AUTH:
	case CL_SENT_CH_RESPONSE:
		para_fd_set(ct->scc.fd, &s->rfds, &s->max_fileno);
		return;

	case CL_RECEIVED_WELCOME:
	case CL_RECEIVED_PROCEED:
	case CL_RECEIVED_CHALLENGE:
		para_fd_set(ct->scc.fd, &s->wfds, &s->max_fileno);
		return;

	case CL_SENDING:
		if (ct->btrn[1]) {
			ret = btr_node_status(ct->btrn[1], 0, BTR_NT_LEAF);
			if (ret < 0)
				sched_min_delay(s);
			else if (ret > 0)
				para_fd_set(ct->scc.fd, &s->wfds, &s->max_fileno);
		}
		/* fall though */
	case CL_EXECUTING:
		if (ct->btrn[0]) {
			ret = btr_node_status(ct->btrn[0], 0, BTR_NT_ROOT);
			if (ret < 0)
				sched_min_delay(s);
			else if (ret > 0)
				para_fd_set(ct->scc.fd, &s->rfds, &s->max_fileno);
		}
		return;
	}
}

static int send_sb(struct client_task *ct, int channel, void *buf, size_t numbytes,
		enum sb_designator band, bool dont_free)
{
	int ret, fd = ct->scc.fd;
	struct iovec iov[2];

	if (!ct->sbc[channel]) {
		struct sb_buffer sbb;
		sb_transformation trafo = ct->status < CL_RECEIVED_PROCEED?
			NULL : sc_trafo;
		sbb = (typeof(sbb))SBB_INIT(band, buf, numbytes);
		ct->sbc[channel] = sb_new_send(&sbb, dont_free, trafo, ct->scc.send);
	}
	ret = sb_get_send_buffers(ct->sbc[channel], iov);
	ret = xwritev(fd, iov, ret);
	if (ret < 0) {
		sb_free(ct->sbc[channel]);
		ct->sbc[channel] = NULL;
		return ret;
	}
	if (sb_sent(ct->sbc[channel], ret)) {
		ct->sbc[channel] = NULL;
		return 1;
	}
	return 0;
}

static int recv_sb(struct client_task *ct, fd_set *rfds,
		struct sb_buffer *result)
{
	int ret;
	size_t n;
	sb_transformation trafo;
	void *trafo_context;
	struct iovec iov;

	if (!FD_ISSET(ct->scc.fd, rfds))
		return 0;
	if (ct->status < CL_SENT_CH_RESPONSE)
		trafo = trafo_context = NULL;
	else {
		trafo = sc_trafo;
		trafo_context = ct->scc.recv;
	}
	if (!ct->sbc[0])
		ct->sbc[0] = sb_new_recv(0, trafo, trafo_context);
again:
	sb_get_recv_buffer(ct->sbc[0], &iov);
	ret = read_nonblock(ct->scc.fd, iov.iov_base, iov.iov_len, rfds, &n);
	if (ret < 0) {
		sb_free(ct->sbc[0]);
		ct->sbc[0] = NULL;
		return ret;
	}
	if (n == 0)
		return 0;
	ret = sb_received(ct->sbc[0], n, result);
	if (ret < 0)
		return ret;
	if (ret == 0)
		goto again;
	ct->sbc[0] = NULL;
	return 1;
}


static char **parse_features(char *buf)
{
	int i;
	const char id[] = "\nFeatures: ";
	char *p, *q, **features;

	p = strstr(buf, id);
	if (!p)
		return NULL;
	p += strlen(id);
	q = strchr(p, '\n');
	if (!q)
		return NULL;
	*q = '\0';
	create_argv(p, ",", &features);
	for (i = 0; features[i]; i++)
		PARA_INFO_LOG("server feature: %s\n", features[i]);
	return features;
}

static int dispatch_sbb(struct client_task *ct, struct sb_buffer *sbb)
{
	int ret;
	const char *designator[] = {SB_DESIGNATORS_ARRAY};

	if (!sbb)
		return 0;
	if (sbb->band < NUM_SB_DESIGNATORS)
		PARA_DEBUG_LOG("band: %s\n", designator[sbb->band]);

	switch (sbb->band) {
	case SBD_AWAITING_DATA:
		ct->status = CL_SENDING;
		ret = 1;
		goto out;
	case SBD_OUTPUT:
		if (iov_valid(&sbb->iov))
			btr_add_output(sbb->iov.iov_base, sbb->iov.iov_len,
				ct->btrn[0]);
		ret = 1;
		goto out;
	case SBD_DEBUG_LOG:
	case SBD_INFO_LOG:
	case SBD_NOTICE_LOG:
	case SBD_WARNING_LOG:
	case SBD_ERROR_LOG:
	case SBD_CRIT_LOG:
	case SBD_EMERG_LOG:
		if (iov_valid(&sbb->iov)) {
			int ll = sbb->band - SBD_DEBUG_LOG;
			para_log(ll, "remote: %s", (char *)sbb->iov.iov_base);
		}
		ret = 1;
		goto deallocate;
	case SBD_EXIT__SUCCESS:
		ret = -E_SERVER_CMD_SUCCESS;
		goto deallocate;
	case SBD_EXIT__FAILURE:
		ret = -E_SERVER_CMD_FAILURE;
		goto deallocate;
	default:
		PARA_ERROR_LOG("invalid band %d\n", sbb->band);
		ret = -E_BAD_BAND;
		goto deallocate;
	}
deallocate:
	free(sbb->iov.iov_base);
out:
	sbb->iov.iov_base = NULL;
	return ret;
}

static bool has_feature(const char *feature, struct client_task *ct)
{
	return find_arg(feature, ct->features) >= 0? true : false;
}

static int send_sb_command(struct client_task *ct)
{
	int i;
	char *command, *p;
	size_t len = 0;

	if (ct->sbc[1])
		return send_sb(ct, 0, NULL, 0, 0, false);

	for (i = 0; i < ct->conf.inputs_num; i++)
		len += strlen(ct->conf.inputs[i]) + 1;
	p = command = para_malloc(len);
	for (i = 0; i < ct->conf.inputs_num; i++) {
		strcpy(p, ct->conf.inputs[i]);
		p += strlen(ct->conf.inputs[i]) + 1;
	}
	PARA_DEBUG_LOG("--> %s\n", command);
	return send_sb(ct, 0, command, len, SBD_COMMAND, false);
}

/*
 * The post select hook for client commands.
 *
 * Depending on the current state of the connection and the status of the read
 * and write fd sets of s, this function performs the necessary steps to
 * authenticate the connection, to send the command given by t->private_data
 * and to receive para_server's output, if any.
 */
static int client_post_select(struct sched *s, void *context)
{
	struct client_task *ct = context;
	int ret = 0;
	size_t n;
	char buf[CLIENT_BUFSIZE];

	ret = task_get_notification(ct->task);
	if (ret < 0)
		goto out;
	if (ct->scc.fd < 0)
		return 0;
	switch (ct->status) {
	case CL_CONNECTED: /* receive welcome message */
		ret = read_nonblock(ct->scc.fd, buf, sizeof(buf), &s->rfds, &n);
		if (ret < 0 || n == 0)
			goto out;
		ct->features = parse_features(buf);
		ct->status = CL_RECEIVED_WELCOME;
		return 0;
	case CL_RECEIVED_WELCOME: /* send auth command */
		if (!FD_ISSET(ct->scc.fd, &s->wfds))
			return 0;
		sprintf(buf, AUTH_REQUEST_MSG "%s sideband%s", ct->user,
			has_feature("aes_ctr128", ct)? ",aes_ctr128" : "");
		PARA_INFO_LOG("--> %s\n", buf);
		ret = write_buffer(ct->scc.fd, buf);
		if (ret < 0)
			goto out;
		ct->status = CL_SENT_AUTH;
		return 0;
	case CL_SENT_AUTH:
		/*
		 * Receive challenge and session keys, decrypt the challenge and
		 * send back the hash of the decrypted challenge.
		 */
		{
		/* decrypted challenge/session key buffer */
		unsigned char crypt_buf[1024];
		struct sb_buffer sbb;
		bool use_aes;

		ret = recv_sb(ct, &s->rfds, &sbb);
		if (ret <= 0)
			goto out;
		if (sbb.band != SBD_CHALLENGE) {
			ret = -E_BAD_BAND;
			free(sbb.iov.iov_base);
				goto out;
		}
		n = sbb.iov.iov_len;
		PARA_INFO_LOG("<-- [challenge] (%zu bytes)\n", n);
		ret = priv_decrypt(ct->key_file, crypt_buf,
			sbb.iov.iov_base, n);
		free(sbb.iov.iov_base);
		if (ret < 0)
			goto out;
		ct->challenge_hash = para_malloc(HASH_SIZE);
		hash_function((char *)crypt_buf, CHALLENGE_SIZE, ct->challenge_hash);
		use_aes = has_feature("aes_ctr128", ct);
		ct->scc.send = sc_new(crypt_buf + CHALLENGE_SIZE, SESSION_KEY_LEN, use_aes);
		ct->scc.recv = sc_new(crypt_buf + CHALLENGE_SIZE + SESSION_KEY_LEN,
			SESSION_KEY_LEN, use_aes);
		hash_to_asc(ct->challenge_hash, buf);
		PARA_INFO_LOG("--> %s\n", buf);
		ct->status = CL_RECEIVED_CHALLENGE;
		return 0;
		}
	case CL_RECEIVED_CHALLENGE:
		ret = send_sb(ct, 0, ct->challenge_hash, HASH_SIZE,
			SBD_CHALLENGE_RESPONSE, false);
		if (ret != 0)
			ct->challenge_hash = NULL;
		if (ret <= 0)
			goto out;
		ct->status = CL_SENT_CH_RESPONSE;
		goto out;
	case CL_SENT_CH_RESPONSE: /* read server response */
		{
		struct sb_buffer sbb;
		ret = recv_sb(ct, &s->rfds, &sbb);
		if (ret <= 0)
			goto out;
		free(sbb.iov.iov_base);
		if (sbb.band != SBD_PROCEED)
			ret = -E_BAD_BAND;
		else
			ct->status = CL_RECEIVED_PROCEED;
		goto out;
		}
	case CL_RECEIVED_PROCEED: /* concat args and send command */
		{
		if (!FD_ISSET(ct->scc.fd, &s->wfds))
			return 0;
		ret = send_sb_command(ct);
		if (ret <= 0)
			goto out;
		ct->status = CL_EXECUTING;
		return 0;
		}
	case CL_SENDING:
		if (ct->btrn[1]) {
			char *buf2;
			size_t sz;
			ret = btr_node_status(ct->btrn[1], 0, BTR_NT_LEAF);
			if (ret == -E_BTR_EOF) {
				/* empty blob data packet indicates EOF */
				PARA_INFO_LOG("blob sent\n");
				ret = send_sb(ct, 1, NULL, 0, SBD_BLOB_DATA, true);
				if (ret >= 0)
					ret = -E_BTR_EOF;
			}
			if (ret < 0)
				goto close1;
			if (ret > 0 && FD_ISSET(ct->scc.fd, &s->wfds)) {
				sz = btr_next_buffer(ct->btrn[1], &buf2);
				assert(sz);
				ret = send_sb(ct, 1, buf2, sz, SBD_BLOB_DATA, true);
				if (ret < 0)
					goto close1;
				if (ret > 0)
					btr_consume(ct->btrn[1], sz);
			}
		}
		/* fall through */
	case CL_EXECUTING:
		if (ct->btrn[0]) {
			ret = btr_node_status(ct->btrn[0], 0, BTR_NT_ROOT);
			if (ret < 0)
				goto close0;
			if (ret > 0 && FD_ISSET(ct->scc.fd, &s->rfds)) {
				struct sb_buffer sbb;
				ret = recv_sb(ct, &s->rfds, &sbb);
				if (ret < 0)
					goto close0;
				if (ret > 0) {
					ret = dispatch_sbb(ct, &sbb);
					if (ret < 0)
						goto close0;
				}
			}
		}
		ret = 0;
		goto out;
	}
close1:
	PARA_INFO_LOG("channel 1: %s\n", para_strerror(-ret));
	btr_remove_node(&ct->btrn[1]);
	if (ct->btrn[0])
		return 0;
	goto out;
close0:
	PARA_INFO_LOG("channel 0: %s\n", para_strerror(-ret));
	btr_remove_node(&ct->btrn[0]);
	if (ct->btrn[1] && ct->status == CL_SENDING)
		return 0;
out:
	if (ret >= 0)
		return 0;
	btr_remove_node(&ct->btrn[0]);
	btr_remove_node(&ct->btrn[1]);
	if (ret != -E_SERVER_CMD_SUCCESS && ret != -E_SERVER_CMD_FAILURE)
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
	if (ct->scc.fd >= 0) {
		close(ct->scc.fd);
		ct->scc.fd = -1;
	}
	free_argv(ct->features);
	ct->features = NULL;
	sc_free(ct->scc.recv);
	ct->scc.recv = NULL;
	sc_free(ct->scc.send);
	ct->scc.send = NULL;
	return ret;
}

/**
 * Connect to para_server and register the client task.
 *
 * \param ct The initialized client task structure.
 * \param s The scheduler instance to register the client task to.
 * \param parent The parent node of the client btr node.
 * \param child The child node of the client node.
 *
 * The client task structure given by \a ct  must be allocated and initialized
 * by \ref client_parse_config() before this function is called.
 *
 * \return Standard.
 */
int client_connect(struct client_task *ct, struct sched *s,
		struct btr_node *parent, struct btr_node *child)
{
	int ret;

	PARA_NOTICE_LOG("connecting %s:%d\n", ct->conf.hostname_arg,
		ct->conf.server_port_arg);
	ct->scc.fd = -1;
	ret = para_connect_simple(IPPROTO_TCP, ct->conf.hostname_arg,
					       ct->conf.server_port_arg);
	if (ret < 0)
		return ret;
	ct->scc.fd = ret;
	ret = mark_fd_nonblocking(ct->scc.fd);
	if (ret < 0)
		goto err_out;
	ct->status = CL_CONNECTED;
	ct->btrn[0] = btr_new_node(&(struct btr_node_description)
		EMBRACE(.name = "client recv", .parent = NULL, .child = child));
	ct->btrn[1] = btr_new_node(&(struct btr_node_description)
		EMBRACE(.name = "client send", .parent = parent, .child = NULL));

	ct->task = task_register(&(struct task_info) {
		.name = "client",
		.pre_select = client_pre_select,
		.post_select = client_post_select,
		.context = ct,
	}, s);
	return 1;
err_out:
	close(ct->scc.fd);
	ct->scc.fd = -1;
	return ret;
}

__noreturn static void print_help_and_die(struct client_task *ct)
{
	struct ggo_help h = DEFINE_GGO_HELP(client);
	bool d = ct->conf.detailed_help_given;

	ggo_print_help(&h, d? GPH_STANDARD_FLAGS_DETAILED : GPH_STANDARD_FLAGS);
	exit(0);
}

/**
 * Parse a client configuration.
 *
 * \param argc Usual argument count.
 * \param argv Usual argument vector.
 * \param ct_ptr Filled in by this function.
 * \param loglevel If not \p NULL, the number of the loglevel is stored here.
 *
 * This checks the command line options given by \a argc and \a argv, sets
 * default values for the user name and the name of the rsa key file and reads
 * further options from the config file.
 *
 * Upon successful return, \a ct_ptr points to a dynamically allocated and
 * initialized client task struct.
 *
 * \return The number of non-option arguments in \a argc/argv on success,
 * negative on errors.
 */
int client_parse_config(int argc, char *argv[], struct client_task **ct_ptr,
		int *loglevel)
{
	char *home = para_homedir();
	int ret;
	struct client_task *ct = para_calloc(sizeof(struct client_task));

	*ct_ptr = ct;
	ct->scc.fd = -1;
	ret = -E_CLIENT_SYNTAX;
	if (client_cmdline_parser(argc, argv, &ct->conf))
		goto out;
	version_handle_flag("client", ct->conf.version_given);
	if (ct->conf.help_given || ct->conf.detailed_help_given)
		print_help_and_die(ct);

	ct->config_file = ct->conf.config_file_given?
		para_strdup(ct->conf.config_file_arg) :
		make_message("%s/.paraslash/client.conf", home);
	ret = file_exists(ct->config_file);
	if (!ret && ct->conf.config_file_given) {
		ret = -E_NO_CONFIG;
		goto out;
	}
	if (ret) {
		struct client_cmdline_parser_params params = {
			.override = 0,
			.initialize = 0,
			.check_required = 0,
			.check_ambiguity = 0,
			.print_errors = 0
		};
		ret = -E_BAD_CONFIG;
		if (client_cmdline_parser_config_file(ct->config_file,
			&ct->conf, &params))
			goto out;
	}
	ct->user = ct->conf.user_given?
		para_strdup(ct->conf.user_arg) : para_logname();

	if (ct->conf.key_file_given)
		ct->key_file = para_strdup(ct->conf.key_file_arg);
	else {
		ct->key_file = make_message("%s/.paraslash/key.%s",
			home, ct->user);
		if (!file_exists(ct->key_file)) {
			free(ct->key_file);
			ct->key_file = make_message("%s/.ssh/id_rsa", home);
		}
	}

	if (loglevel)
		*loglevel = get_loglevel_by_name(ct->conf.loglevel_arg);
	PARA_INFO_LOG("loglevel: %s\n", ct->conf.loglevel_arg);
	PARA_INFO_LOG("config_file: %s\n", ct->config_file);
	PARA_INFO_LOG("key_file: %s\n", ct->key_file);
	ret = ct->conf.inputs_num;
out:
	free(home);
	if (ret < 0) {
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
		client_close(ct);
		*ct_ptr = NULL;
	}
	return ret;
}

/**
 * Parse the client configuration and open a connection to para_server.
 *
 * \param argc See \ref client_parse_config.
 * \param argv See \ref client_parse_config.
 * \param ct_ptr See \ref client_parse_config.
 * \param loglevel See \ref client_parse_config.
 * \param parent See \ref client_connect().
 * \param child See \ref client_connect().
 * \param sched See \ref client_connect().
 *
 * This function combines client_parse_config() and client_connect(). It is
 * considered a syntax error if no command was given, i.e. if the number
 * of non-option arguments is zero.
 *
 * \return Standard.
 */
int client_open(int argc, char *argv[], struct client_task **ct_ptr,
		int *loglevel, struct btr_node *parent, struct btr_node *child,
		struct sched *sched)
{
	int ret = client_parse_config(argc, argv, ct_ptr, loglevel);

	if (ret < 0)
		return ret;
	if (ret == 0) {
		ret = -E_CLIENT_SYNTAX;
		goto fail;
	}
	ret = client_connect(*ct_ptr, sched, parent, child);
	if (ret < 0)
		goto fail;
	return 1;
fail:
	client_close(*ct_ptr);
	*ct_ptr = NULL;
	return ret;
}
