/*
 * Copyright (C) 2006 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file write_common.c common functions of para_audiod and para_write */

#include <regex.h>

#include "para.h"
#include "string.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "buffer_tree.h"
#include "write.h"
#include "error.h"
#include "write_common.h"

/** the array containing the names of all supported writers */
const char *writer_names[] ={WRITER_NAMES};

/** the array of supported writers */
struct writer writers[NUM_SUPPORTED_WRITERS] = {WRITER_ARRAY};

/**
 * Call the init function of each supported paraslash writer.
 */
void writer_init(void)
{
	int i;

	FOR_EACH_WRITER(i)
		writers[i].init(&writers[i]);
}
/**
 * Check if given string is a valid command line for any writer.
 *
 * \param \wa String of the form writer_name:options.
 * \param writer_num Contains the number of the writer upon success.
 *
 * This function checks whether \a wa starts with the name of a supported
 * paraslash writer, optionally followed by a colon and any options for that
 * writer.  If a valid writer name was found and further are present, the
 * remaining part of \a wa is passed to that writer's config parser.
 *
 * \return On success, a pointer to the gengetopt args info struct is returned
 * and \a writer_num contains the number of the writer. Otherwise this function
 * prints an error message and calls exit().
 */
void *check_writer_arg_or_die(const char *wa, int *writer_num)
{
	int i, ret, argc;
	const char *cmdline;
	char **argv;
	void *conf;

	if (!wa || !*wa) {
		i = DEFAULT_WRITER;
		cmdline = NULL;
		goto check;
	}
	PARA_INFO_LOG("checking %s\n", wa);
	FOR_EACH_WRITER(i) {
		const char *name = writer_names[i];
		size_t len = strlen(name);
		char c;

		if (strlen(wa) < len)
			continue;
		if (strncmp(name, wa, len))
			continue;
		c = wa[len];
		if (!c || c == ' ') {
			cmdline = c? wa + len + 1 : NULL;
			goto check;
		}
	}
	PARA_EMERG_LOG("invalid writer %s\n", wa);
	exit(EXIT_FAILURE);
check:
	ret = create_shifted_argv(cmdline, " \t", &argv);
	if (ret < 0) {
		PARA_EMERG_LOG("%s: %s\n", wa, para_strerror(-ret));
		exit(EXIT_FAILURE);
	}
	argc = ret;
	argv[0] = make_message("%s_write", writer_names[i]);
	*writer_num = i;
	conf = writers[i].parse_config_or_die(argc, argv);
	free_argv(argv);
	return conf;
}

/**
 * Open a writer node and register the corresponding task.
 *
 * \param wn The writer node to open.
 * \param parent The parent btr node (the source for the writer node).
 * \param s The scheduler instance to register the task to.
 *
 * The configuration of the writer node stored in \p wn->conf must be
 * initialized before this function may be called.
 */
void register_writer_node(struct writer_node *wn, struct btr_node *parent,
		struct sched *s)
{
	struct writer *w = writers + wn->writer_num;

	wn->btrn = btr_new_node(&(struct btr_node_description)
		EMBRACE(.name = writer_names[wn->writer_num], .parent = parent,
		.handler = w->execute, .context = wn));
	wn->task = task_register(&(struct task_info) {
		.name = writer_names[wn->writer_num],
		.pre_select = w->pre_select,
		.post_select = w->post_select,
		.context = wn,
	}, s);
}

/**
 * Print the help text of all writers to stdout.
 *
 * \param flags Passed to \ref ggo_print_help().
 */
void print_writer_helps(unsigned flags)
{
	int i;

	printf_or_die("\nAvailable writers: ");
	FOR_EACH_WRITER(i)
		printf_or_die("%s%s", i? " " : "", writer_names[i]);
	printf_or_die("\n");
	FOR_EACH_WRITER(i) {
		struct writer *w = writers + i;

		if (!w->help.short_help)
			continue;
		printf_or_die("\n%s: %s", writer_names[i],
			w->help.purpose);
		ggo_print_help(&w->help, flags);
	}
}

static void get_btr_value(struct btr_node *btrn, const char *cmd,
		int32_t *result)
{
	char *buf = NULL;
	int ret = btr_exec_up(btrn, cmd, &buf);

	if (ret < 0) {
		/*
		 * This really should not happen. It means one of our parent
		 * nodes died unexpectedly. Proceed with fingers crossed.
		 */
		PARA_CRIT_LOG("cmd %s: %s\n", cmd, para_strerror(-ret));
		*result = 0;
		return;
	}
	ret = para_atoi32(buf, result);
	assert(ret >= 0);
	free(buf);
}

/**
 * Ask parent btr nodes for the sample rate of the current stream.
 *
 * \param btrn Where to start the search.
 * \param result Filled in by this function.
 *
 * This function is assumed to succeed and terminates on errors.
 */
void get_btr_sample_rate(struct btr_node *btrn, int32_t *result)
{
	get_btr_value(btrn, "sample_rate", result);
}

/**
 * Ask parent btr nodes for the channel count of the current stream.
 *
 * \param btrn See \ref get_btr_sample_rate.
 * \param result See \ref get_btr_sample_rate.
 */
void get_btr_channels(struct btr_node *btrn, int32_t *result)
{
	get_btr_value(btrn, "channels", result);
}

/**
 * Ask parent btr nodes for the number of bits per sample and the byte sex.
 *
 * \param btrn See \ref get_btr_sample_rate.
 * \param result Contains the sample format as an enum sample_format type.
 */
void get_btr_sample_format(struct btr_node *btrn, int32_t *result)
{
	get_btr_value(btrn, "sample_format", result);
}
